// TCR Decompressor  v1.04
//
// (c) UK  Andrew Giddings  2003
// gidds@cix.co.uk
// See http://www.cix.co.uk/~gidds/Software/TCR.html for the latest version.
//
// DESCRIPTION
//
// TCR is a format for compressed text files.  It usually achieves a 40-50%
// reduction, which isn't quite as good as LZ-based compressors such as Zip,
// but unlike those, TCR files can be decompressed starting at any point.  This
// makes it particularly suitable for reader programs, which can display the
// compressed file directly without uncompressing it all first - such programs
// as VReader and EBook on the Psion Series 5, and Reader on the Psion Series 3.
//
// This decompressor is written in standard, platform-neutral C (well, gcc).
//
// The TCR compression format is easy to describe: after the fixed header is a
// dictionary of 256 strings, each preceded by a length byte.  The rest of the
// file is a list of codes from this dictionary.
//
// This is my own work, not derived from Ian Young's or Barry Childress' work
// (although of course it'll be very similar as it's the same file format).
//
// LICENCE
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public Licence as published by the
// Free Software Foundation; either version 2 of the Licence, or (at your
// option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public Licence
// for more details.  (http://www.gnu.org/licenses/gpl.txt)
//
// (That said, I'm happy to answer questions and hear comments (anatomical or
// otherwise), and would appreciate hearing of any improvements you make.)
//
// HISTORY
//
// v1.00
//	First version.
//
// v1.01
//	Handle zero codes.  Although my compressor doesn't use the zero
//	code, some other compressors may.
//
// v1.02
//	Use internal buffers for speed.  Instead of relying on the stdio package to
//	buffer the i/o, we now decompress from one large buffer to another, and
//	handle file reads and writes as necessary.  This is much more complicated,
//	but it runs about four times as fast!  Also tighten up checks for dodgy
//	files.
//
// v1.03
//	No changes, but best to keep the version numbers consistent...
//
// v1.04
//  Fixed all compilation warnings.  GPLed.


////////////////////////////////////////////////////////////////////////////////
// Headers

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <fcntl.h>
#include <stdarg.h>


////////////////////////////////////////////////////////////////////////////////
// Constants

const char *COPYRIGHT = "TCR Decompressor  v1.04   (c) UK  Andrew Giddings  2003";
const char *FILE_HEADER = "!!8-Bit!!";


////////////////////////////////////////////////////////////////////////////////
// Prototypes

void doFile(char *inName);
void decode(FILE *inFile, FILE *outFile);
void die(const char *format,...);


////////////////////////////////////////////////////////////////////////////////

int main(int argc, char *argv[])
{
	fprintf(stderr, "%s\n", COPYRIGHT);

	// If no params, filter stdin to stdout
	if (argc == 1)
		decode(stdin, stdout);
	else
		while (--argc)
			doFile(*++argv);

	exit(EXIT_SUCCESS);
}


////////////////////////////////////////////////////////////////////////////////

void doFile(char *inName)
{	char outName[PATH_MAX], *s;
	FILE *in, *out;

	strcpy(outName, inName);
	s = outName + strlen(outName) - 4;
	if (!strcasecmp(s, ".tcr"))
		strcpy(s, ".txt");
	else
		strcat(outName, ".txt");

	if (access(outName, F_OK) == 0)
		die("%s already exists", outName);

	errno = 0;

	if (!(in = fopen(inName, "rb")))
		die("Can't open %s - %s", inName, strerror(errno));

	if (!(out = fopen(outName, "wb")))
		die("Can't create %s - %s", outName, strerror(errno));

	printf("Decompressing %s into %s...\n", inName, outName);

	decode(in, out);

	fclose(in);
	fclose(out);
}


////////////////////////////////////////////////////////////////////////////////

#define BUFSIZE 65536

void decode(FILE *inFile, FILE *outFile)
{	unsigned char *dict[256], *inBuf, *outBuf, *in, *out, *inEnd, *outEnd;
	int c, i, dictLen[256];

	// Check header

	inBuf = (char *) malloc(BUFSIZE);
	if (!inBuf)
		die("Not enough memory for input buffer");
	outBuf = (char *) malloc(BUFSIZE);
	if (!outBuf)
		die("Not enough memory for output buffer");

	errno = 0;

	fread(inBuf, 1, strlen(FILE_HEADER), inFile);
	if (errno)
		die("Can't read header - %s", strerror(errno));

	if (strncmp(inBuf, FILE_HEADER, strlen(FILE_HEADER)))
		die("Not a TCR compressed file");

	// Read dictionary

	for (i = 0; i < 256; ++i)
	{	if ((dictLen[i] = getc(inFile)) < 0)
			die("Can't read dictionary entry %i length - %s", i, strerror(errno));
		if (dictLen[i] > 255)
			die("Invalid dictionary entry %i length %i - %s", i, dictLen[i],
					strerror(errno));
		if (!(dict[i] = (char *) malloc(dictLen[i])))
			die("Out of memory for entry %i", i);
		if (fread(dict[i], 1, dictLen[i], inFile) < dictLen[i])
		{	if (errno)
				die("Can't read dictionary entry %i - %s", i, strerror(errno));
			else
				die("Malformed TCR file - too short");
		}
	}

	// Decode file

	in = inEnd = inBuf;
	out = outBuf;
	outEnd = outBuf + BUFSIZE;

	for (;;)
	{	if (in >= inEnd)
		{	int inLen;
			inLen = fread(inBuf, 1, BUFSIZE, inFile);
			if (inLen < 0)
				die("Can't read - %s", strerror(errno));
			else if (inLen == 0)
				break;
			in = inBuf;
			inEnd = inBuf + inLen;
		}

		c = *in++;
		if (dictLen[c] == 0)
		{	printf("Warning: unused code %i found...\n", c);
			continue;
		}

		if (out + dictLen[c] > outEnd)
		{	if (fwrite(outBuf, 1, out - outBuf, outFile) < out - outBuf)
				die("Can't write - %s", strerror(errno));
			out = outBuf;
		}
		memcpy(out, dict[c], dictLen[c]);
		out += dictLen[c];
	}

	if (fwrite(outBuf, 1, out - outBuf, outFile) < out - outBuf)
		die("Can't write - %s", strerror(errno));

	// Free dictionary

	for (i = 255; i >= 0; --i)
		free(dict[i]);
	free(outBuf);
	free(inBuf);
}


////////////////////////////////////////////////////////////////////////////////
// Handle a fatal error

void die(const char *format,...)
{	va_list list;

	va_start(list, format);
	vfprintf(stderr, format, list);
	fprintf(stderr, "!\n");
	exit(EXIT_FAILURE);
}
